/*
** Module   : FramePlacementManager.java
** Abstract : The class handles proper placement of frames at the screen. 
**
** Copyright (c) 2005-2012, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- ------------------Description-------------------
** 001 EVL 20050713   @21692 Creation based on general concept. Necessary
**                           modification to get compiled.
** 002 EVL 20050715   @21705 Changing the location of the widget types
**                           constants.
** 003 EVL 20050715   @21706 Adding handling of the pseudo widgets - Skip
**                           and Space.
** 004 EVL 20050720   @21800 Changes to meet the new XML formats. And
**                           taking into account the frame creation flags.
** 005 EVL 20050723   @21807 Adding additional handling for selection
**                           list and combo box widgets.
** 006 EVL 20050725   @21811 Adding FillIn and Label reconstruction
**                           methods.
** 007 EVL 20050726   @21858 Handling parent-child frames relationship.
** 008 EVL 20050727   @21889 Taking into account the different data types
**                           of the down frames.
** 009 EVL 20050728   @21912 Changing base for decimal creation to initial
**                           string value for down frame.
** 010 EVL 20050728   @21913 Adding different types of the fill-in widget
**                           processing.
** 011 EVL 20050728   @21914 Adding layout processing to the frame
**                           reconstruction.
** 012 EVL 20050731   @21928 Adding the ability to create the undefined
**                           logical variable for the FillIn widget.
** 013 EVL 20050731   @21927 The method to reconstruct the text widget is
**                           added to the process of reconstruction.
** 014 EVL 20050822   @22199 Unification for the P2JFrame/P2JContainer
**                           interfaces.
** 015 EVL 20050824   @22239 Adding the events/triggers processing.
** 016 EVL 20050825   @22274 Adding methods to get widget ID.
** 017 EVL 20050828   @22374 Changing event handler initialization to
**                           set up only those events which we really
**                           look for.
** 018 EVL 20050830   @22415 Adding focus related messages ENTRY and 
**                           LEAVE.
** 019 SIY 20050831   @22512 Refactoring (see @@22394-22409) 
** 020 EVL 20050903   @22456 Changing the logic of handling event list in
**                           config classes. Adding key event handler.
**                           (merged changes from EVL).
** 021 EVL 20050908   @22616 Moving event related string constants to the
**                           P2JEvent class.
** 022 EVL 20050912   @22681 Adding validation as another event to handle
** 023 EVL 20050913   @22700 Splitting the VALIDATE and LEAVE events in
**                           two independent cases. Fixing the focus
**                           handling issue for selection-list widget.
** 024 EVL 20050919   @22749 Changing the target of the key events for
**                           some widgets. Selection list is touched by 
**                           this.
** 025 EVL 20050922   @22814 Adding editing interceptor.
** 026 SIY 20051226   @24121 Refactoring and changes for the new approach.
** 027 NVS 20060125   @24068 Removed all traces of DownFrame class.
** 028 SIY 20060128   @24122 Added frame manager.
** 029 SIY 20060129   @24151 Minor cleanups.      
** 030 SIY 20060201   @24209 Javadoc warning cleanup.     
** 031 SIY 20060203   @24241 Fixed centered frame placement. Added method 
**                           to hide unused frames.
** 032 SIY 20060209   @24449 Updated to accommodate changes in xxxConfig
**                           classes. Minor cleanups.
** 033 SIY 20060213   @24577 Added support for the DialogBox; fixes for 
**                           the frame hiding.
** 034 SIY 20060315   @25065 Yet another try to fix frame hiding.
** 035 SIY 20060316   @25077 Workaround for the application main menu.
** 036 SIY 20060322   @25213 Helper methods for down frames.
** 037 SIY 20060331   @25293 Apply component configuration via containing 
**                           frame.
** 038 SIY 20060424   @25687 hideUnusedFrames() now returns boolean flag
**                           which allows to track if any frames were 
**                           actually hidden.
** 039 GES 20060502   @25927 Remove top-level temp text elements when
**                           hiding frames.
** 040 SIY 20060428   @25985 Added EDITOR widget.
** 041 SIY 20060504   @25985 Attempt to fix frame hiding issue.
** 042 SIY 20060509   @26040 Fixed frame hiding when it is requested from 
**                           hideAll().
** 043 GES 20060510   @26046 hideAll() needs to clear all temp text
**                           widgets (put screen output is all removed).
**                           Added new method removeTemps().
** 044 SIY 20060510   @26064 Fixed a regression in frame hiding.
** 045 GES 20060510   @26067 Safety code to ensure that TempText removal
**                           will work using absolute (screen coordinates)
**                           of the frame being hidden.
** 046 SIY 20060511   @26074 More fixes in hiding/pausing logic.
** 047 SIY 20060525   @26548 Fixed placement of the DOWN(0) frames.
** 048 NVS 20060526   @26583 Placement of frame on screen is recalculated
**                           unless it is realized.
** 049 SIY 20060526   @26913 New approach in pause handling.
** 050 SIY 20060607   @26962 Do not hide frames with noHide flag set.
** 051 NVS 20060608   @27017 getComponent() method has been expanded to
**                           cover the session window component.
** 052 SIY 20060609   @27099 Added helper methods for handling out of 
**                           scope frames. 
** 053 NVS 20060613   @27171 Added getTopFrame() method to return the
**                           frame on the top of the Z-order.
** 054 NVS 20060613   @27181 getTopFrame() should ignore invisible frames
**                           to remain compatible with the existing code.
** 055 SIY 20060616   @27400 All uses of pauseBeforeHide() made exception 
**                           safe.
** 056 SIY 20060617   @27418 Hiding of one frame ignores focus flags now.
** 057 SIY 20060621   @27531 More changes in frame hiding algorithms.
** 058 SIY 20060628   @27656 Support for dynamic down frames.
** 059 GES 20060714   @28063 getTopFrame() must ignore frames that are
**                           out of scope (marked as isCanHide() == true)
**                           or disabled.
** 060 SIY 20060726   @28231 Refactoring for the new hiding approach.
** 061 GES 20060802   @28300 Minor performance improvements.
** 062 GES 20060803   @28412 Cleanup and refactoring.
** 063 SIY 20060807   @28453 Changed hideFrames() signature to minimize
**                           unnecessary hiding/viewing of frames.
**                           Some minor cleanups.
** 064 SIY 20060824   @28862 Removed references to DialogBox.
** 065 SIY 20061001   @30044 Fixed frame placement issue which caused an 
**                           unexpected frame hiding in some cases.
** 066 SIY 20061009   @30228 Improved calculation of start row.
** 067 SIY 20061010   @30269 Minor code cleanup.
** 068 NVS 20061020   @30585 Added tracking of frame rectangle affected by
**                           overlay frames.
** 069 GES 20061027   @30780 Changed detection of inactive frames to use
**                           Frame.isInScope() instead of Frame.isAlive()
**                           (which did not work and no longer exists).
** 070 NVS 20061101   @30866 frame can't be placed at a negative row even
**                           if its height is larger than screen (happened
**                           in reports).
** 071 NVS 20061107   @30995 Fixed bugs in placeFrame() not always hiding
**                           properly other frames.
** 072 GES 20061107   @30998 Removed some dead code.
** 073 NVS 20061109   @31046 Fixed regression caused by the fix 071.
** 074 SIY 20061116   @31292 Fixed handling of overlay frames.
** 075 GES 20061120   @31331 Removed dead code, cleaned up code style and
**                           format.  Fixed frame removal to remove from
**                           Window's frame list at the same time the
**                           frame and its components are removed from
**                           the master widget list.  This was a major
**                           memory leak.  Added cleanupDeadFrames().
** 076 SIY 20061128   @31428 Moved findFrames() to Window as more 
**                           appropriate place.
** 077 SIY 20070117   @31846 Minor change in removeFrame()
** 078 SIY 20070131   @32030 Adjust DOWN frame even if its position is 
**                           specified.
** 079 SIY 20070309   @32382 Removed unnecessary cleanupDeadFrames method.
** 080 SIY 20070316   @32417 Minor cleanup.
** 081 SIY 20070327   @32642 Make sure that current frame will not trigger
**                           pausing in hideFrames().
** 082 NVS 20070404   @32761 Added new method getFrameIDs() which returns
**                           array of all frame IDs in their Z-order.
** 083 SIY 20070412   @32954 Fixed one specific case of automatic frame 
**                           placement.
** 084 SIY 20070420   @33155 Fixed one specific case of automatic frame
**                           placement of overlay frames.
** 085 SIY 20070501   @33323 Fixed removing of components. Old approach in 
**                           some cases (browse columns) leaved widgets in 
**                           the list which resulted to layout distortions
**                           when widget IDs interfered.
** 086 SIY 20070501   @33335 Removed redundant widget adding code. 
** 087 SIY 20070501   @34559 Fix for placement of frame over another frame 
*                            when underlying frame will be removed.
*                            Minor performance enhancement in toArray()
*                            calls. 
** 088 SIY 20070805   @34807 Use current terminal dimensions when placing
**                           frames.
** 089 NVS 20070815   @34850 Optimized component removal loop.
** 090 NVS 20071025   @35521 Added new method getFrameIdForStream() which
**                           searches the list of frames for one connected
**                           to the specified stream.
** 091 SIY 20071031   @35647 Fixed frame placement in certain conditions.
**                           Ref: #35646
** 092 SIY 20071108   @35756 Added call to Frame.postprocess().
**                           Ref: #35755
** 093 SIY 20071120   @35929 Support for Frame 'need pause' flag 
**                           maintenance.
**                           Ref: #35928
** 094 SVL 20071206   @36224 Make widgets disabled by default.
** 095 SIY 20071214   @36360 Fix for regression caused by #093.
**                           Ref: #36348
** 096 SIY 20071214   @36387 Fixed frame placement issue.
**                           Ref: #36347                            
** 097 NVS 20071229   @36613 Removed obsolete getFrameIdForStream().
** 098 SIY 20080117   @36819 Call adjustDown() for all down frames, not 
**                           just dynamic ones. This is necessary because 
**                           initial fixed frame height can be greater than
**                           screen height, but how it needs to be adjusted
**                           we can determine only when frame real position
**                           is assigned.
** 099 SIY 20080129   @36931 Moved processing of TempText to Frame.
**                           Ref: #36929
** 100 SIY 20080218   @37113 Moved doLayout() call to Frame.postprocess()
**                           Ref: #37107
** 101 CA  20080226   @37197 placeFrame() must hide only the frames with
**                           the same redirection status as the frame being
**                           placed. Ref: #37196
** 102 SIY 20080304   @37291 Changed getTopFrame() signature in order to
**                           allow unfiltered search.
**                           Ref: #37288
** 103 SIY 20080306   @37316 Added helper method resetStartRow().
**                           Ref: #37313
** 104 SIY 20080310   @37421 Hide streamed and interactive frames 
**                           independently. Also, when searching frames 
**                           which fit into rectangle for frame placement
**                           check space up to right side of the screen.
**                           Ref: #37420
** 105 NVS 20080320   @37613 Added supporting code for maintaining a list
**                           of displayed frames and methods to access the
**                           list. This list is required in waitForWorker
**                           when choosing the initial focus.
** 106 SIY 20080324   @37654 Fixed regressions caused by #104. Note that we
**                           don't check space up to right side of screen
**                           anymore. Instead we relaxed requirements for
**                           "garbage collecting" and it resolves the issue
**                           for which "wide" rectangle was used.
** 107 NVS 20080328   @37726 Backed out part of #105 as it didn't do well.
**                           The separate list of displayed frames is no
**                           longer maintained. Two lookup methods added
**                           with #104, getFirstVisibleFrame() and
**                           getNextVisibleFrame() are reimplemented using
**                           the list of top level components maintained
**                           in Window.
** 108 NVS 20080402   @37789 Modified getFirstVisibleFrame() to return the
**                           top dialog box in case there is any, in 
**                           preference to the regular frames.
** 109 SIY 20080415   @37988 Fixed placement of overlay frames.
**                           Ref: #37987
** 110 SIY 20080605   @38573 Avoid changing widget enabled state without a
**                           real need in order to avoid repainting.
** 111 NVS 20080605   @38578 Removed code maintaining overlay rectangle
**                           for frames in placeFrame(). It's no longer 
**                           needed.
** 112 SIY 20080609   @38650 Added getFramesAbove() helper method.
** 113 SIY 20080610   @38673 Improved getFramesAbove() to let it collect
**                           only those frames which overlap current.
** 114 SIY 20080715   @39114 Renamed method.
** 115 SIY 20080729   @39220 Fixed frame placement for case when small 
**                           frames are placed so they don't intersect 
**                           initially but affect placement of frames at 
**                           next iteration.
** 116 CA  20080902   @39748 Dialogs intersected by this frame are not 
**                           hidden by placeFrame(Frame). hideDialogs() was
**                           modified to receive a list of frames which
**                           will not be hidden.
** 117 CA  20081014   @40085 When hiding an out-of-scope frame, it must
**                           be cleared first and a pause must be 
**                           triggered.
** 118 SIY 20081015   @40090 Limit #117 to DOWN frames only to avoid 
**                           regressions.
** 119 SIY 20081016   @40099 Limit #117 to DOWN frames which are not 
**                           cleared yet to avoid regressions.
** 120 SIY 20081017   @40156 Finally rolled back #117 and replaced it with
**                           other fix (see ThinClient#471).
** 121 CA  20081124   @40683 Only frames marked with canHide=true can hide
**                           other frames. If, there is an exception during
**                           the PAUSE triggered before hiding other frames,
**                           the frames must not be hidden.
** 122 CA  20081127   @40704 Modified H121 to use the 'canHideOthers' flag.
** 123 SIY 20081218   @40956 Added support for placement quirk mode.                               
** 124 SIY 20081222   @40967 Allow turn off placement quirk mode at current
**                           trigger level nesting.                               
** 125 SIY 20081224   @40993 Reworked #124.
** 126 CA  20090107   @41023 getTopLeveComponents was changed to package
**                           private.
** 127 NVS 20090213   @41315 Dead code removal.
** 128 SIY 20090305   @41413 Explicit placement of first frame into last 
**                           available row breaks placement of remaining
**                           frames. Fixed start row calculation for frames
**                           with specified placement row which hide other
**                           frames during placement.
** 129 SIY 20090310   @41506 Placement of visible frame should mark all 
**                           frames which are hidden during placement as
**                           hidden explicitly.
** 130 SVL 20090402   @41717 Perform explicit frame hiding even if user
**                           has pressed F4.
** 131 CA  20090422   @41839 Added methods to retrieve the subsequent or the
**                           precedent enabled frame, relative to a specified 
**                           frame.
** 132 SIY 20090526   @42463 Fixed detection of hiding of different instances
**                           of the same logical frame.
** 133 SIY 20090601   @42507 Slightly changed logic of different instances of
**                           logical frames detection. 
** 134 SIY 20090810   @43579 Added detection of hiding frame at the same row as
**                           given one in hideFrames().
** 135 SIY 20091118   @44428 Added addFrame() method.
** 136 CA  20091208   @44492 Fixed a case of HIDE for overlay frames - this
**                           should not change the starting drawing position 
**                           for the next viewed/displayed frame.
** 137 SIY 20091218   @44518 Fixed one special case of placement of frames 
**                           when empty frames present at the screen.
** 138 SIY 20100104   @44521 Fixed frame placement after frame hiding.
** 139 SIY 20100125   @44550 Temporarily disabled #137 to avoid regressions.
** 140 SIY 20100204   @44565 Reimplemented #137 fix using different approach.
** 141 EVL 20100713          The hideFrames() method is changed by adding
**                           condition to exception handler. This way we can
**                           consider F4 EndCondition the same way as normal
**                           exit by space bar and consume the exception.
**                           This does fix some UI painting issues. Also new
**                           import statement is now added to the top of
**                           the file.
** 142 EVL 20100720          The fix #141 improvements. Some logic need the
**                           exception to be propagated when we press F4 to
**                           hide the frame. Now the final section execution
**                           will be forced to process completely.
** 143 SIY 20100619          Switching to own implementation of character UI.
** 144 SIY 20100804          Changes in imports, some refactorings.
** 145 SIY 20100822          Refactored AttributeContainer support.
** 146 SIY 20100831          Changes related to refactoring of FrameHelper.
** 147 SIY 20100915          Changes related to refactoring of ThinClient and
**                           Frame.
** 148 SIY 20101015          Removed widget registry (moved to separate class).
**                           Renamed from Screen to FramePlacementManager.
** 149 SIY 20111101          Updated imports.
** 150 SIY 20111115          Added setting of physical location.
** 151 CS  20121031          Removed client specific references.
*/

package com.goldencode.p2j.ui.chui;

import java.util.*;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.widget.*;
import com.goldencode.p2j.util.*;

/**
 * The class handles proper placement of frames at the screen.
 */
public class FramePlacementManager
{
   /** hideFrames call nesting */
   private int hideNesting = 0;
   
   /** Start row for the frame placement. */
   private int startRow = 0;
  
   /** Track activation/deactivation of placement quirk. */
   private boolean placementQuirk = false;

   /** Reusable frame comparator. */
   private FrameComparator comparator = new FrameComparator();
   
   /**
    * Place new frame at the screen. Hide other frames if necessary.
    * 
    * @param   frame
    *          Frame to place.
    */
   public void placeFrame(Frame frame)
   {
      boolean redirected = frame.isRedirected();

      FrameConfig cfg = frame.config();
      ArrayList<Frame> list = new ArrayList<Frame>();
      
      Dimension d = OutputManager.instance().screenDimension();
      d.height -= Window.RESERVED_LINES;
      
      int lastRow = d.height;
      
      Point pos = new Point(BaseConfig.INV_COORD, BaseConfig.INV_COORD);
      int newRow = startRow;

      if (!frame.isAutoplacing() || frame.isVisible())
         pos = frame.location();

      if (frame.isDialog())
      {  
         if (pos.x == BaseConfig.INV_COORD)
            pos.x = (d.width - frame.width())/2;
         if (pos.y == BaseConfig.INV_COORD)
            pos.y = (d.height - frame.height())/2;
         setFrameLocation(frame, pos);
         return;
      }
      
      if (pos.x == BaseConfig.INV_COORD)
      {
         pos.x = 0;
         
         if (cfg.isCentered())
         {
            pos.x = (d.width - frame.width())/2; 
         }
      }

      if (pos.y == BaseConfig.INV_COORD) 
      {
         if (frame.isCanHideOthers())
         {
            hideMarkedFrames(OutputManager.instance().isRedirected());
         }
         newRow = startRow;

         if (placementQuirk)
         {
            // force placement of the frame at the row 0 
            pos.y = 0;
         }
      }

      Dimension frameDim = frame.dimension();
      
      if (pos.y == BaseConfig.INV_COORD)
      {
         int height = frame.getMinHeight();
         
         // do automatic placement
         // find the place for the new frame
         if ((newRow + height) > lastRow || cfg.isOverlay())
         {
            int row = lastRow - height;
            
            if (row < 0)
               row = 0;
            
            Window.findFrames(list, 
                              new Rectangle(new Point(pos.x, row), 
                                            new Dimension(frame.width(), 
                                                          height)), 
                              true, 
                              redirected);
            
            Frame[] frames = list.toArray(new Frame[list.size()]);
            
            // sort frames by ROW
            Arrays.sort(frames, comparator);

            if (!cfg.isOverlay() || (row + height) > lastRow)
            {
               if (frames.length > 0 && row >= frames[0].location().y)
               {
                  row = frames[0].location().y;
               }
               
               
               {
                  // In some rare cases between topmost frame which fits in
                  // the rectangle and bottommost frame which does not fit 
                  // into rectangle can be some free space. Take this situation
                  // into account and adjust row accordingly.
                  ArrayList<Frame> topList = new ArrayList<Frame>();
                  Rectangle topRect = new Rectangle(0, 0, row - 1, frame.width());
                  Window.findFrames(topList, topRect, true, redirected);
                  
                  if (topList.size() > 0)
                  {
                     frames = topList.toArray(new Frame[topList.size()]);
                     
                     // sort frames by ROW
                     Arrays.sort(frames, comparator);
                     
                     int bottomRow = 
                        frames[frames.length - 1].location().y +
                        frames[frames.length - 1].height();
                     
                     if (bottomRow < row)
                        row = bottomRow;
                  }
               }
            }
            else if (cfg.isOverlay() && row > 0)
            {
               // overlay frames with free ROW tend to move themselves up
               // as much as possible unless they find free space or int
               ArrayList<Frame> list2 = new ArrayList<Frame>();
               
               pos.y = row;
               
               Dimension tmpDim = frameDim.copy(); 
               
               do
               {
                  list2.clear();
                  row = pos.y;
                  pos.y--;
                  tmpDim.height++;
                  
                  Window.findFrames(list2, 
                                    new Rectangle(pos, tmpDim), 
                                    true,
                                    redirected);
               }
               while (pos.y >= 0 && list.equals(list2));
               
               if (!frame.isDynamic())
               {
                  pos.y = row;
                  list.clear();
                  Window.findFrames(list, 
                                    new Rectangle(pos, frameDim), 
                                    true,
                                    redirected);
               }
            }
            
            pos.y = row;
         }
         else
         {
            pos.y = newRow;

            Window.findFrames(list, 
                              new Rectangle(pos, frameDim), 
                              true,
                              redirected);
            
            if (list.size() != 0)
            {
               newRow = pos.y;
               
               while(list.size() != 0)
               {
                  pos.y++;
                  
                  if (pos.y + height > lastRow)
                  {
                     pos.y = Math.min(list.get(0).location().y,
                                      newRow);
                     
                     Window.findFrames(list,
                                       new Rectangle(pos, frameDim),
                                       true,
                                       redirected);
                     break;
                  }
                  
                  list.clear();
                  Window.findFrames(list, 
                                    new Rectangle(pos, frameDim), 
                                    true,
                                    redirected);
                  
                  if (list.size() == 0)
                  {
                     newRow = pos.y;
                     break;
                  }
               }
            }
         }

         if (list.size() == 0)
         {
            newRow = pos.y;
            
            while(pos.y > 0)
            {
               pos.y--;
               
               Window.findFrames(list, 
                                 new Rectangle(pos, frameDim), 
                                 true,
                                 redirected,
                                 true);
               
               if (list.size() == 0)
               {
                  newRow = pos.y;
                  continue;
               }
               
               list.clear();
               pos.y = newRow;
               break;
            }
         }

         setFrameLocation(frame, pos);
         
         if (frame.isDown())
            frame.adjustDown();
         
         newRow = pos.y + frame.height();
      }
      else
      {
         if (placementQuirk)
            frameDim.width = d.width - pos.x;
         
         // frame is with specified position
         Window.findFrames(list, 
                           new Rectangle(pos, frameDim), 
                           true, 
                           redirected);
         
         setFrameLocation(frame, pos);
         
         if (frame.isDynamic())
            frame.adjustDown();
         
         if (!cfg.isOverlay())
         {
            newRow = getNewRow(newRow, pos.y + frame.height(), list);
         }

         // if calculated row does not fit the screen, then next frame will
         // be placed starting from bottom of current frame
         if (newRow >= lastRow)
            newRow = pos.y + frame.height();
         
         // special case which enables 'placement quirk' mode
         if (pos.y == lastRow)
         {
            placementQuirk = true;
         }
         else if (UiUtils.getTopVisibleFrame() == null && newRow == lastRow)
         {
            newRow = 0;
         }
      }
      
      list.remove(frame);
      
      try
      {
         if (list.size() > 0)
         {
            // remove all dialogs from the list, they are not considered as
            // overlaid, if the frame is behind them
            Iterator<Frame> iter = list.iterator();
            while (iter.hasNext())
            {
               Frame f = iter.next();
               if (f.isDialog())
               {
                  iter.remove();
               }
            }

            Frame[] frl = list.toArray(new Frame[list.size()]);
            boolean ovl = cfg.isOverlay();
            int saveStartRow = startRow;

            try
            {
               if (frame.isVisible())
               {
                  for (Frame item : frl)
                  {
                     item.setExplicitHide(true);
                  }
               }
              
               hideFrames(frl, true, !ovl);
            }
            finally
            {
               startRow = saveStartRow;

               if (frl.length > 0 && pos.y != BaseConfig.INV_COORD)
               {
                  newRow = pos.y + frame.height();
               }
            }
         }
      }
      finally
      {
         // calculate vertical position for the next frame
         list.clear();
         list.add(frame);
         
         Window.findFrames(list, 
                           new Rectangle(new Point(0,0), d), 
                           true, 
                           redirected);
         
         Frame[] frames = list.toArray(new Frame[list.size()]);
         
         Arrays.sort(frames, comparator);

         int bottomRow = 
            frames[frames.length - 1].location().y +
            frames[frames.length - 1].height();
         
         if (startRow != bottomRow || newRow > startRow)
            startRow = newRow;
      }
   }

   /**
    * Calculate and set frame physical location.
    * 
    * @param   frame
    *          Target frame.
    * @param   pos
    *          Location in character cells.
    */
   private void setFrameLocation(Frame frame, Point pos)
   {
      frame.setLocation(pos);
      int x = pos.x * FontManager.getFontWidth(frame.config().getFont());
      int y = pos.y * FontManager.getFontHeight(frame.config().getFont());
      frame.setPhysicalLocation(x, y);
   }

   /**
    * Hide all frames.
    * 
    * @param   doPause
    *          If <code>true</code> then pause will be handled as necessary.
    */
   public void hideAllFrames(boolean doPause)
   {
      Widget[] c = UiUtils.getTopLevelComponents();
      
      hideFrames(c, doPause, true);
   }

   /**
    * Hide single frame.
    * 
    * @param   frame
    *          Frame to hide.
    * @param   doPause
    *          If <code>true</code> then pause will be handled if necessary.
    */
   public void hideFrame(Frame frame, boolean doPause)
   {
      if (!frame.isVisible())
         return;
      
      // add a restoration mechanism for the startRow field, as if this is an 
      // overlay frame it should not affect the position for the next frame
      // draw
      int saveStartRow = startRow;
      
      try
      {
         Point loc = frame.location();
         
         if (loc.y < startRow)
         {
            startRow = loc.y;
         }
         
         hideFrames(new Frame[] {frame}, doPause, true);
         frame.resetPlacementQuirk(true);
      }
      finally
      {
         // if this is a overlay frame, then the drawing position should not
         // be changed
         if (frame.config().isOverlay())
         {
            startRow = saveStartRow;
         }
      }
   }
   
   /**
    * Hide frames from specified array of components.
    * 
    * @param   list
    *          Array of Component instances.
    * @param   doPause
    *          If <code>true</code> pause will be handled if necessary.
    * @param   doHide
    *          Perform real hiding if <code>true</code> and simulate
    *          otherwise.
    */
   private void hideFrames(Widget[] list, boolean doPause, boolean doHide)
   {
      hideNesting++;
      
      if (hideNesting == 1 && doPause)
      {
         doPause = false;
         
         for (int i = 0; i < list.length; i++)
         {
            if (!list[i].isVisible())
               continue;

            if (list[i] instanceof Frame)
            {
               Frame frm = (Frame) list[i];

               if (frm.isDialog())
                  continue;

               if (frm.isNeedPause())
                  doPause = true;
            }
         }
      }
      else
         doPause = false;

      // determine whether it is an implicit or explicit hiding
      boolean explicit = true;

      for (int i = 0; doHide && i < list.length; i++)
      {
         if (list[i] instanceof Frame)
         {
            explicit = ((Frame) list[i]).isExplicitHide();
            break;
         }
      }

      boolean error = false;
      try
      {
         if (doPause)
            ThinClient.getInstance().pauseBeforeHide();
      }
      catch (RuntimeException e)
      {
         error = true;
         
         // In the case of pressing F4 as the exception cause we still
         // need the frame hiding to be done anyway
         if (e instanceof EndConditionException)
            explicit = true;
         
         throw e;
      }
      finally
      {
         // Perform explicit frame hiding even if user has pressed F4.
         if (!error || explicit)
         {
            for (int i = 0; doHide && i < list.length; i++)
            {
               if (!list[i].isVisible())
                  continue;

               if (list[i] instanceof Frame)
               {
                  Frame frm = (Frame) list[i];
                  Point loc = frm.location();

                  // the drawing start row must be changed only if this is not
                  // an overlay frame
                  if (loc.y < startRow && !frm.config().isOverlay())
                     startRow = loc.y;

                  frm.setVisible(false);
               }
            }
         }

         hideNesting--;
      }
   }

   /**
    * Calculate new value for start row for frame placement.
    * 
    * @param   current
    *          Current value of the start row.
    * @param   proposed
    *          Proposed value for start row.
    * @param   list
    *          List of frames which will be deleted. 
    * 
    * @return  Calculated start row value. 
    */
   private static int getNewRow(int current, int proposed, ArrayList<Frame> list)
   {
      if (current < proposed) // increasing of the value is always allowed
         return proposed;

      current = proposed;
      
      Widget[] c = UiUtils.getTopLevelComponents();

      for (Widget component : c)
      {
         if (component instanceof Frame)
         {
            Frame frm = (Frame) component;
            FrameConfig cfg = frm.config();
            
            if (!frm.isVisible() || cfg.isOverlay() || !list.contains(frm))
               continue;

            if (frm.location().y + frm.height() > current)
               current = frm.location().y + frm.height();
         }
      }
      
      return current;
   }
   
   /**
    * Hide frames which can be hidden.
    * 
    * @param   streamed
    *          <code>true</code> hide streamed frames, <code>false</code> -
    *          hide frames from interactive terminal.
    */
   private void hideMarkedFrames(boolean streamed)
   {
      Widget[] c = UiUtils.getTopLevelComponents();

      for (int i = 0; i < c.length; i++)
      {
         if (c[i] instanceof Frame)
         {
            Frame frm = (Frame) c[i];
            FrameConfig cfg = frm.config();
            
            if (!frm.isVisible()  || 
                 cfg.isNoHide()   || 
                 cfg.isOverlay()  || 
                 frm.isDialog()   ||
                 !frm.isCanHide() ||
                 frm.isStreamed() != streamed)
            {
               continue;
            }
            
            if (startRow > frm.location().y)
               startRow = frm.location().y;
            
            frm.setVisible(false);
         }
      }
   }

   /**
    * Reset initial row for frame placement.
    */
   public void resetStartRow()
   {
      startRow = 0;
   }   
   
   /**
    * Reset placement quick mode.
    */
   public void resetPlacementQuirk()
   {
      this.placementQuirk = false;
   }

   /**
    * Local comparator for two Frame instances.
    */
   private static class FrameComparator
   implements Comparator<Frame>
   {
      /**
       * Compare two frames for the row.
       * 
       * @see java.util.Comparator#compare(java.lang.Object, java.lang.Object)
       */
      @Override
      public int compare(Frame o1, Frame o2)
      {
         return o1.location().y - o2.location().y;
      }
   }
}
